@@ -32,8 +32,474 @@
 #define MIN(a, b) (((a) > (b)) ? (b) : (a))
 #endif
 
+typedef struct GcContext {
+   long time;
+} GcContext;
+
+static float lut_unused_minutes_limit = 5.0;
+
+static int lut_info_level = 0;
+
+#define _LUT_LOG(level, ...) do{\
+     if (level <= lut_info_level)\
+       fprintf (stdout, __VA_ARGS__);\
+     fflush(NULL);\
+     }while(0)
+#define LUT_LOG(...) _LUT_LOG(1, __VA_ARGS__)
+#define LUT_INFO(...) _LUT_LOG(2, __VA_ARGS__)
+#define LUT_DETAIL(...) _LUT_LOG(3, __VA_ARGS__)
+
+static int gc_fishes (Babl *babl, void *userdata)
+{
+  GcContext *context = userdata;
+  if (babl->class_type == BABL_FISH_PATH)
+  {
+    if (babl->fish_path.u8_lut)
+    {
+      if (context->time - babl->fish_path.last_lut_use >
+          1000 * 1000 * 60 * lut_unused_minutes_limit)
+      {
+        void *lut =babl->fish_path.u8_lut;
+        BABL(babl)->fish_path.u8_lut = NULL;
+        free (lut);
+        BABL(babl)->fish.pixels = 0;
+        LUT_LOG("freeing LUT %s to %s unused for >%.1f minutes\n",
+                babl_get_name (babl->conversion.source),
+                babl_get_name (babl->conversion.destination),
+                lut_unused_minutes_limit);
+      }
+      else if (lut_info_level >=4)
+      {
+        LUT_DETAIL("active LUT %s to %s  %8li pixels last used %.1f minutes ago\n",
+                babl_get_name (babl->conversion.source),
+                babl_get_name (babl->conversion.destination),
+                babl->fish.pixels,
+         (context->time - babl->fish_path.last_lut_use)/1000.0/1000.0/60.0);
+      }
+    }
+    else if (lut_info_level >= 4 && babl->fish.pixels)
+    {
+        if (babl->fish_path.is_u8_color_conv)
+        LUT_DETAIL("potential LUT %s to %s  %8li pixels\n",
+                babl_get_name (babl->conversion.source),
+                babl_get_name (babl->conversion.destination),
+                babl->fish.pixels);
+        else if (lut_info_level >=5)
+        LUT_DETAIL("%i step path %s to %s  %8li pixels\n",
+                babl->fish_path.conversion_list->count,
+                babl_get_name (babl->conversion.source),
+                babl_get_name (babl->conversion.destination),
+                babl->fish.pixels);
+    }
+    babl->fish.pixels /= 2; // decay pixel count// this is enough that we *will* reach 0
+  }
+  return 0;
+}
+                           
+static void
+babl_gc_fishes (void)
+{
+  GcContext context;
+  context.time = babl_ticks ();
+  if (lut_info_level >= 5)
+  {
+     fprintf (stdout, "\e[H\e[2J");
+  }
+  babl_fish_class_for_each (gc_fishes, &context);
+  //malloc_trim (0); 
+  //  is responsibility of higher layers
+}
+
+#define BABL_LIKELY(x)      __builtin_expect(!!(x), 1)
+#define BABL_UNLIKELY(x)    __builtin_expect(!!(x), 0)
+
+static float timings[256] = {0,};
+
+static inline void _do_lut (uint32_t *lut,
+                           int   source_bpp,
+                           int   dest_bpp,
+                           const void *__restrict__ source,
+                           void *__restrict__ destination,
+                           long n)
+{
+        if (source_bpp == 4 && dest_bpp == 16)
+        {
+          uint32_t *src = (uint32_t*)source;
+          uint32_t *dst = (uint32_t*)destination;
+          while (n--)
+          {
+             uint32_t col = *src++;
+             uint32_t lut_offset = col & 0xffffff;
+             float alpha = (col>>24)/255.0;
+
+             *dst++ = lut[lut_offset*4+0];
+             *dst++ = lut[lut_offset*4+1];
+             *dst++ = lut[lut_offset*4+2];
+             ((float*)(dst))[0] = alpha;
+             dst++;
+          }
+        }
+        else if (source_bpp == 4 && dest_bpp == 8)
+        {
+          uint32_t *src = (uint32_t*)source;
+          uint16_t *dst = (uint16_t*)destination;
+          uint16_t *lut16 = (uint16_t*)lut;
+          while (n--)
+          {
+             uint32_t col = *src++;
+             uint32_t lut_offset = col & 0xffffff;
+             uint16_t alpha = (col>>24) << 8;
+
+             dst[0] = lut16[lut_offset*2+0];
+             dst[1] = lut16[lut_offset*2+1];
+             dst[2] = lut16[lut_offset*2+2];
+             dst[3] = alpha;
+             dst+=4;
+          }
+        }
+        else if (source_bpp == 2 && dest_bpp == 16)
+        {
+          uint16_t *src = (uint16_t*)source;
+          uint32_t *dst = (uint32_t*)destination;
+          while (n--)
+          {
+             uint32_t col = *src++;
+             *dst++ = lut[col*4+0];
+             *dst++ = lut[col*4+1];
+             *dst++ = lut[col*4+2];
+             *dst++ = lut[col*4+3];
+          }
+        }
+        else if (source_bpp == 4 && dest_bpp == 4)
+        {
+          uint32_t *src = (uint32_t*)source;
+          uint32_t *dst = (uint32_t*)destination;
+          while (n--)
+          {
+             uint32_t col = *src++;
+             *dst = col & 0xff000000;
+             *dst |= lut[col & 0xffffff];
+             dst++;
+          }
+        }
+        else if (source_bpp == 2 && dest_bpp == 4)
+        {
+          uint16_t *src = (uint16_t*)source;
+          uint32_t *dst = (uint32_t*)destination;
+          while (n--)
+          {
+             *dst = lut[*src++];
+             dst++;
+          }
+        }
+        else if (source_bpp == 2 && dest_bpp == 2)
+        {
+          uint16_t *src = (uint16_t*)source;
+          uint16_t *dst = (uint16_t*)destination;
+          uint16_t *lut16 = (uint16_t*)lut;
+          while (n--)
+          {
+             *dst = lut16[*src++];
+             dst++;
+          }
+        }
+        else if (source_bpp == 1 && dest_bpp == 4)
+        {
+          uint8_t *src = (uint8_t*)source;
+          uint32_t *dst = (uint32_t*)destination;
+          while (n--)
+          {
+             *dst = lut[*src++];
+             dst++;
+          }
+        }
+        else if (source_bpp == 3 && dest_bpp == 3)
+        {
+          uint8_t *src = (uint8_t*)source;
+          uint8_t *dst = (uint8_t*)destination;
+          while (n--)
+          {
+             uint32_t col = src[0]*256*256+src[1]*256+src[2];
+             uint32_t val = lut[col];
+             dst[2]=(val >> 16) & 0xff;
+             dst[1]=(val >> 8) & 0xff;
+             dst[0]=val & 0xff;
+             dst+=3;
+             src+=3;
+          }
+        }
+        else if (source_bpp == 3 && dest_bpp == 4)
+        {
+          uint8_t *src = (uint8_t*)source;
+          uint32_t *dst = (uint32_t*)destination;
+          while (n--)
+          {
+             uint32_t col = src[0]*256*256+src[1]*256+src[2];
+             *dst = lut[col];
+             dst++;
+             src+=3;
+          }
+        }
+}
+
+
+void babl_test_lut (uint32_t *lut,
+             int   source_bpp,
+             int   dest_bpp,
+             void *__restrict__ source,
+             void *__restrict__ dest,
+             long count);
+void babl_test_lut (uint32_t *lut,
+             int   source_bpp,
+             int   dest_bpp,
+             void *__restrict__ source,
+             void *__restrict__ dest,
+             long count)
+{
+   _do_lut (lut, source_bpp, dest_bpp, source, dest, count);
+}
+
+static inline float lut_timing_for (int source_bpp, int dest_bpp)
+{
+  return timings[source_bpp * 16 + dest_bpp];
+}
+
+static void measure_timings(void)
+{
+   int num_pixels = babl_get_num_path_test_pixels () * 1000;
+   int pairs[][2]={{4,4},{4,8},{3,4},{3,3},{2,4},{2,2},{1,4},{2,16},{4,16}};
+   uint32_t *lut = malloc (256 * 256 * 256 * 16);
+   uint32_t *src = malloc (num_pixels * 16);
+   uint32_t *dst = malloc (num_pixels * 16);
+
+   memset (lut, 11, 256 * 256 * 256 *16);
+   memset (src, 12, num_pixels * 16);
+
+   if (getenv ("BABL_LUT_INFO"))
+   {
+      lut_info_level = atoi (getenv ("BABL_LUT_INFO"));
+   }
+   if (getenv ("BABL_LUT_UNUSED_LIMIT"))
+   {
+      lut_unused_minutes_limit = atof (getenv ("BABL_LUT_UNUSED_LIMIT"));
+   }
+
+   LUT_LOG("BABL_LUT_UNUSED_LIMIT=%.1f\n", lut_unused_minutes_limit);
+
+   LUT_LOG("measuring lut timings          \n");
+   for (size_t p = 0; p < sizeof (pairs)/sizeof(pairs[0]);p++)
+   {
+     int source_bpp = pairs[p][0];
+     int dest_bpp = pairs[p][1];
+     long start,end;
+     start = babl_ticks ();
+     babl_test_lut (lut, source_bpp, dest_bpp, src, dst, num_pixels);
+
+     end = babl_ticks ();
+     timings[source_bpp * 16 + dest_bpp] = (end-start)/1000.0;
+       LUT_LOG ("   %ibpp to %ibpp: %.2f\n", source_bpp, dest_bpp,
+          timings[source_bpp * 16 + dest_bpp]
+                     );
+   }
+   free (lut);
+   free (src);
+   free (dst);
+}
+
+static inline void
+process_conversion_path (BablList   *path,
+                         const void *source_buffer,
+                         int         source_bpp,
+                         void       *destination_buffer,
+                         int         dest_bpp,
+                         long        n);
+
+static inline int babl_fish_lut_process_maybe (const Babl *babl,
+                                               const char *source,
+                                               char *destination,
+                                               long        n,
+                                               void       *data)
+{
+     int source_bpp = babl->fish_path.source_bpp;
+     int dest_bpp = babl->fish_path.dest_bpp;
+     uint32_t *lut = (uint32_t*)babl->fish_path.u8_lut;
+
+
+     if (BABL_UNLIKELY(!lut && babl->fish.pixels >= 128 * 256))
+     {
+       LUT_LOG("generating LUT for %s to %s\n",
+               babl_get_name (babl->conversion.source),
+               babl_get_name (babl->conversion.destination));
+       if (source_bpp ==4 && dest_bpp == 4)
+       {
+         lut = malloc (256 * 256 * 256 * 4);
+         for (int o = 0; o < 256 * 256 * 256; o++)
+           lut[o] = o;
+         process_conversion_path (babl->fish_path.conversion_list,
+                                  lut, 4,
+                                  lut, 4,
+                                  256*256*256);
+         for (int o = 0; o < 256 * 256 * 256; o++)
+           lut[o] = lut[o] & 0x00ffffff;
+       }
+       else if (source_bpp == 4 && dest_bpp == 16)
+       {
+         uint32_t *temp_lut = malloc (256 * 256 * 256 * 4);
+         lut = malloc (256 * 256 * 256 * 16);
+         for (int o = 0; o < 256 * 256 * 256; o++)
+           temp_lut[o] = o;
+         process_conversion_path (babl->fish_path.conversion_list,
+                                  temp_lut, 4,
+                                  lut, 16,
+                                  256*256*256);
+         free (temp_lut);
+       }
+       else if (source_bpp == 4 && dest_bpp == 8)
+       {
+         uint32_t *temp_lut = malloc (256 * 256 * 256 * 4);
+         lut = malloc (256 * 256 * 256 * 8);
+         for (int o = 0; o < 256 * 256 * 256; o++)
+           temp_lut[o] = o;
+         process_conversion_path (babl->fish_path.conversion_list,
+                                  temp_lut, 4,
+                                  lut, 8,
+                                  256*256*256);
+         free (temp_lut);
+       }
+       else if (source_bpp == 3 && dest_bpp == 3)
+       {
+         uint8_t *temp_lut = malloc (256 * 256 * 256 * 3);
+         uint8_t *temp_lut2 = malloc (256 * 256 * 256 * 3);
+         int o = 0;
+         lut = malloc (256 * 256 * 256 * 4);
+         for (int r = 0; r < 256; r++)
+         for (int g = 0; g < 256; g++)
+         for (int b = 0; b < 256; b++, o++)
+         {
+           temp_lut[o*3+0]=r;
+           temp_lut[o*3+1]=g;
+           temp_lut[o*3+2]=b;
+         }
+         process_conversion_path (babl->fish_path.conversion_list,
+                                  temp_lut, 3,
+                                  temp_lut2, 3,
+                                  256*256*256);
+         babl_process (babl_fish (babl_format ("R'G'B' u8"), babl_format ("R'G'B'A u8")),
+                       temp_lut2, lut, 256*256*256);
+         for (int o = 0; o < 256 * 256 * 256; o++)
+           lut[o] = lut[o] & 0x00ffffff;
+         free (temp_lut);
+         free (temp_lut2);
+       }
+       else if (source_bpp == 3 && dest_bpp == 4)
+       {
+         uint8_t *temp_lut = malloc (256 * 256 * 256 * 3);
+         int o = 0;
+         lut = malloc (256 * 256 * 256 * 4);
+         for (int r = 0; r < 256; r++)
+         for (int g = 0; g < 256; g++)
+         for (int b = 0; b < 256; b++, o++)
+         {
+           temp_lut[o*3+0]=r;
+           temp_lut[o*3+1]=g;
+           temp_lut[o*3+2]=b;
+         }
+         process_conversion_path (babl->fish_path.conversion_list,
+                                  temp_lut, 3,
+                                  lut, 4,
+                                  256*256*256);
+         for (int o = 0; o < 256 * 256 * 256; o++)
+           lut[o] = lut[o] & 0x00ffffff;
+         free (temp_lut);
+       }
+       else if (source_bpp == 2 && dest_bpp == 2)
+       {
+         uint16_t *temp_lut = malloc (256 * 256 * 2);
+         lut = malloc (256 * 256 * 4);
+         for (int o = 0; o < 256*256; o++)
+         {
+           temp_lut[o]=o;
+         }
+         process_conversion_path (babl->fish_path.conversion_list,
+                                  temp_lut, 2,
+                                  lut, 2,
+                                  256*256);
+         free (temp_lut);
+       }
+       else if (source_bpp == 2 && dest_bpp == 4)
+       {
+         uint16_t *temp_lut = malloc (256 * 256 * 2);
+         lut = malloc (256 * 256 * 4);
+         for (int o = 0; o < 256*256; o++)
+         {
+           temp_lut[o]=o;
+         }
+         process_conversion_path (babl->fish_path.conversion_list,
+                                  temp_lut, 2,
+                                  lut, 4,
+                                  256*256);
+         for (int o = 0; o < 256 * 256; o++)
+           lut[o] = lut[o] & 0x00ffffff;
+         free (temp_lut);
+       }
+       else if (source_bpp == 2 && dest_bpp == 16)
+       {
+         uint16_t *temp_lut = malloc (256 * 256 * 2);
+         lut = malloc (256 * 256 * 16);
+         for (int o = 0; o < 256*256; o++)
+         {
+           temp_lut[o]=o;
+         }
+         process_conversion_path (babl->fish_path.conversion_list,
+                                  temp_lut, 2,
+                                  lut, 16,
+                                  256*256);
+         free (temp_lut);
+       }
+       else if (source_bpp == 1 && dest_bpp == 4)
+       {
+         uint8_t *temp_lut = malloc (256);
+         lut = malloc (256 * 4);
+         for (int o = 0; o < 256; o++)
+         {
+           temp_lut[o]=o;
+         }
+         process_conversion_path (babl->fish_path.conversion_list,
+                                  temp_lut, 1,
+                                  lut, 4,
+                                  256);
+         for (int o = 0; o < 256; o++)
+           lut[o] = lut[o] & 0x00ffffff;
+         free (temp_lut);
+       }
+
+       if (babl->fish_path.u8_lut == NULL)
+       {
+         (BABL(babl)->fish_path.u8_lut) = lut;
+         // XXX need memory barrier?
+         if ((BABL(babl)->fish_path.u8_lut) != lut)
+         {
+           free (lut);
+           lut = babl->fish_path.u8_lut;
+         }
+       }
+       else
+       {
+         free (lut);
+         lut = babl->fish_path.u8_lut;
+       }
+     }
+     if (lut)
+     {
+        _do_lut (lut, source_bpp, dest_bpp, source, destination, n);
+        BABL(babl)->fish_path.last_lut_use = babl_ticks ();
+     }
+     return 0;
+}
+
+
+
 #define MAX_BUFFER_SIZE            512
-#define ITERATIONS                 4
 
 int   babl_in_fish_path = 0;
 
@@ -100,7 +566,6 @@ _babl_fish_create_name (char       *buf,
 static int max_path_length (void);
 
 static int debug_conversions = 0;
-int _babl_instrument = 0;
 
 double
 _babl_legal_error (void)
@@ -123,12 +588,6 @@ _babl_legal_error (void)
   else
     debug_conversions = 0;
 
-  env = getenv ("BABL_INSTRUMENT");
-  if (env && env[0] != '\0')
-    _babl_instrument = 1;
-  else
-    _babl_instrument = 0;
-
   return error;
 }
 
@@ -145,9 +604,9 @@ max_path_length (void)
   if (env)
     max_length = atoi (env);
   else
-    max_length = 2; /* reducing this number makes finding short fishes much
+    max_length = 3; /* reducing this number makes finding short fishes much
                        faster - even if we lose out on some of the fast
-                       bigger fish, the fishes we can get with a max_length of 2
+                       bigger fish, the fishes we can get with a max_length of 3
                        is actually 5, since we deepen the search to that
                        depth if none are found within two steps in the
                        initial search.
@@ -317,6 +776,9 @@ int
 _babl_fish_path_destroy (void *data)
 {
   Babl *babl=data;
+  if (babl->fish_path.u8_lut)
+    free (babl->fish_path.u8_lut);
+  babl->fish_path.u8_lut = NULL;
   if (babl->fish_path.conversion_list)
     babl_free (babl->fish_path.conversion_list);
   babl->fish_path.conversion_list = NULL;
@@ -472,6 +934,55 @@ _babl_fish_prepare_bpp (Babl *babl)
        default:
          babl_log ("-eeek{%i}\n", babl_dest->instance.class_type - BABL_MAGIC);
      }
+
+  {
+  int source_bpp = babl->fish_path.source_bpp;
+  int dest_bpp = babl->fish_path.dest_bpp;
+  if (//source->format.space != destination->format.space &&
+     (
+        (source_bpp == 2 && dest_bpp == 16)
+      ||(source_bpp == 4 && dest_bpp == 16)
+      ||(source_bpp == 4 && dest_bpp == 4)
+      ||(source_bpp == 4 && dest_bpp == 8)
+      ||(source_bpp == 3 && dest_bpp == 4)
+      ||(source_bpp == 2 && dest_bpp == 4)
+      ||(source_bpp == 2 && dest_bpp == 2)
+      ||(source_bpp == 1 && dest_bpp == 4)
+      ||(source_bpp == 3 && dest_bpp == 3)
+      )
+     )
+  {
+     // as long as the highest 8bit of the 32bit of a 4 byte input is ignored
+     // (alpha) - and it is not an associated color model. A 24 bit LUT provides
+     // exact data. Thus this is valid for instance for "YA half"
+
+     if ((babl->conversion.source->format.type[0]->bits < 32 &&
+          (source_bpp < 4 
+         || (babl->conversion.source->format.model->flags & BABL_MODEL_FLAG_ASSOCIATED)==0)))
+     {
+       static int measured_timings = 0;
+       float scaling = 10.0;
+       if (!measured_timings) measure_timings ();
+       measured_timings = 1;
+       LUT_LOG ("%sLUT for %s to %s   %.2f%s%.2f\n",
+
+       ((lut_timing_for (source_bpp, dest_bpp) * scaling) <
+                           babl->fish_path.cost)?"possible ":"no ",
+
+                        babl_get_name (babl->conversion.source),
+                        babl_get_name (babl->conversion.destination),
+                        (lut_timing_for (source_bpp, dest_bpp) * scaling),
+       ((lut_timing_for (source_bpp, dest_bpp) * scaling) <
+                           babl->fish_path.cost)?" < ":" > ",
+                        babl->fish_path.cost);
+       if ((lut_timing_for (source_bpp, dest_bpp) * scaling) <
+                           babl->fish_path.cost)
+       {
+         babl->fish_path.is_u8_color_conv = 1;
+       }
+     }
+  }
+  }
 }
 
 void
@@ -552,6 +1063,7 @@ babl_fish_path2 (const Babl *source,
     static const Babl *run_once[512]={NULL};
     int i;
     int done = 0;
+
     for (i = 0; run_once[i]; i++)
     {
       if (run_once[i] == source->format.space)
@@ -582,7 +1094,6 @@ babl_fish_path2 (const Babl *source,
     {
       babl_conversion_class_for_each (show_item, (void*)source->format.space);
     }
-
   }
 
   babl = babl_calloc (1, sizeof (BablFishPath) +
@@ -601,10 +1112,11 @@ babl_fish_path2 (const Babl *source,
   babl->fish_path.conversion_list = babl_list_init_with_size (BABL_HARD_MAX_PATH_LENGTH);
 
 
+
   {
     PathContext pc;
     int start_depth = max_path_length ();
-    int end_depth = start_depth + 2 + ((destination->format.space != sRGB)?1:0);
+    int end_depth = start_depth + 1 + ((destination->format.space != sRGB)?1:0);
     end_depth = MIN(end_depth, BABL_HARD_MAX_PATH_LENGTH);
 
     pc.current_path = babl_list_init_with_size (BABL_HARD_MAX_PATH_LENGTH);
@@ -652,6 +1164,7 @@ babl_fish_path2 (const Babl *source,
     }
 
   _babl_fish_prepare_bpp (babl);
+
   _babl_fish_rig_dispatch (babl);
   /* Since there is not an already registered instance by the required
    * name, inserting newly created class into database.
@@ -704,6 +1217,24 @@ babl_fish_path_process (const Babl *babl,
                         long        n,
                         void       *data)
 {
+  BABL(babl)->fish.pixels += n;
+  if (babl->fish_path.is_u8_color_conv)
+  {
+     if (babl_fish_lut_process_maybe (babl,
+                                      source, destination, n,
+                                      data))
+       return;
+  }
+  else
+  {
+    static long conv_counter = 0;
+    conv_counter+=n;
+    if (conv_counter > 1000 * 1000 * 10) // run gc every 10 megapixels
+    {
+      conv_counter = 0;
+      babl_gc_fishes ();
+    }
+  }
   process_conversion_path (babl->fish_path.conversion_list,
                            source,
                            babl->fish_path.source_bpp,
@@ -790,8 +1321,6 @@ _babl_process (const Babl *cbabl,
 {
   Babl *babl = (void*)cbabl;
   babl->fish.dispatch (babl, source, destination, n, *babl->fish.data);
-  if (_babl_instrument)
-    babl->fish.pixels += n;
   return n;
 }
 
@@ -823,8 +1352,6 @@ babl_process_rows (const Babl *fish,
   if (n <= 0)
     return 0;
 
-  if (_babl_instrument)
-    babl->fish.pixels += n * rows;
   for (row = 0; row < rows; row++)
     {
       babl->fish.dispatch (babl, (void*)src, (void*)dst, n, *babl->fish.data);
@@ -838,7 +1365,7 @@ babl_process_rows (const Babl *fish,
 #include <stdint.h>
 
 #define BABL_ALIGN 16
-static void inline *align_16 (unsigned char *ret)
+static inline void *align_16 (unsigned char *ret)
 {
   int offset = BABL_ALIGN - ((uintptr_t) ret) % BABL_ALIGN;
   ret = ret + offset;
