@@ -555,6 +555,7 @@ switch (trc->type)
       break;
     }
   case BABL_TRC_FORMULA_SRGB:
+  // fall through
   case BABL_TRC_FORMULA_CIE:
     {
       int lut_size = 512;
@@ -1079,10 +1080,10 @@ babl_space_from_icc (const char   *icc_data,
       }
       break;
     case BABL_ICC_INTENT_ABSOLUTE_COLORIMETRIC:
-      *error = "absolute colormetric not implemented";
+      *error = "absolute colorimetric not implemented";
       break;
     case BABL_ICC_INTENT_SATURATION:
-      *error = "absolute stauration not supported";
+      *error = "saturation not supported";
       break;
   }
 
@@ -1504,8 +1505,11 @@ ConvertUTF16toUTF8 (const UTF16   **sourceStart,
 	}
 	switch (bytesToWrite) { /* note: everything falls through. */
 	    case 4: *--target = (UTF8)((ch | byteMark) & byteMask); ch >>= 6;
+	    // fall through
 	    case 3: *--target = (UTF8)((ch | byteMark) & byteMask); ch >>= 6;
+	    // fall through
 	    case 2: *--target = (UTF8)((ch | byteMark) & byteMask); ch >>= 6;
+	    // fall through
 	    case 1: *--target =  (UTF8)(ch | firstByteMark[bytesToWrite]);
 	}
 	target += bytesToWrite;
